# ===============================================================================================================
#    SrvrPowerCtrl - a plugin for SqueezeCenter 7.1
#      Allows shutdown/restart/suspend/hibernation of your SqueezeCenter server
#      hardware via your SqueezeBox's IR remote or via a SBC
#
#    Version 7.3-RC8-20090505
#
#    Copyright (C) 2008 Gordon Harris
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of the GNU General Public License,
#    version 2, which should be included with this software.
#
#    Based on Adrian C's (aka tenwiseman) Server Power Control plugin: http://adrianonline.wordpress.com/
#                                                   http://forums.slimdevices.com/showthread.php?t=32674
#    Copyright AdrianOnline.Net Feb 2007. Initially modeled on Max Spicer's Max2.pm
#
#    Also includes code from PeterW's AllQuiet plugin: http://www.tux.org/~peterw/slim/slim7/AllQuiet/
#    AllQuiet copyright (c) 2007 by Peter Watkins (peterw@tux.org)
#
#
#    See a duscussion of this plugin at http://forums.slimdevices.com/showthread.php?t=48521
#
# ===============================================================================================================
#
#    Settings.pm -- Settings & prefs routines..
#

package Plugins::SrvrPowerCtrl::Settings;


use strict;
use base qw(Slim::Web::Settings);

use Slim::Utils::Prefs;
use Slim::Utils::Validate;
use Slim::Utils::OSDetect;
use File::Spec::Functions qw(:ALL);

if ($^O eq 'MSWin32') {		## Are we running windows?
	eval {
	require Win32;
	import Win32;
	} ;
}

# Global var hash.  Accessed by all modules in the plugin..

use vars qw(%g);

%g = (
	nVersion				=> 20090505,					#version of this code..
	szOS					=> Slim::Utils::OSDetect::OS(),	#Operating system we're running on..
	aBlockAction			=> [ ],							#Array of block action hashes.. [ ] vs. () is important here to reserve the data space, apparently..
	aActions				=> [ ], 						#Array of action/menu/message/command hash items
	log						=> undef,						#log hash
	prefs					=> undef,						#preferences
	tSPCWatchdogTimer		=> 0,							#our timer ID so we can kill it
	tPendingActionTimer		=> 0,							#timer ID of the pending action...so we can kill it on cancel..
);


#pref names must be in Hungarian notation for generic validation to work.
#pref name prefix == data type and determines type of validation:
#'b' == boolian (i.e. checkbox value)
#'n' == numeric (i.e. int)
#'sz' == string, null terminated

my @prefNames = (	'bInclude_Shutdown',		'bInclude_Shutdown2SN',			'szShutdown_cmd',
					'bInclude_Restart',			'szRestart_cmd',
					'bInclude_Suspend',			'bInclude_Suspend2SN',			'szSuspend_cmd',
					'bInclude_Hibernate', 		'bInclude_Hibernate2SN',		'szHibernate_cmd',
					'bInclude_SCRestart', 		'szSCRestart_cmd',
					'bInclude_WebInterface',
					'bPowerOffPlayers',
					'nRegretDelay',
					'nGoToSNDelay',
					'bNoWebUIRedirect2SN',
					'bGotoSNAllPlayers',		'szPushSNPlayerMACs',
					'bSetRTCWakeForAlarm',		'szSetRTCWake_cmd',				'nRTCWakeupAdvance',
					'bRetrievePlayersFromSN',	'nRetrievePlayersFromSNWait',	'szPullSNPlayerMACs',
					'bHookSleepButton', 		'szSleepButtonAction',
					'bUseSleepEndWatchdog',		'szSleepEndWatchdogAction',
					'bUseIdleWatchdog', 		'nIdleWatchdogTimeout',			'szIdleWatchdogAction',
					'bUseEODWatchdog',			'nEODWatchdogTimeout',			'szEODWatchdogStartTime',		'szEODWatchdogEndTime',	'szEODWatchdogAction',	'szEODWatchdog_cmd',
					'bUseOnWakeupWatchdog',		'szOnWakup_cmd',

		   );

my %prefDefaults = (	'bInclude_Shutdown'				=>	1,
						'bInclude_Shutdown2SN'			=>	1,
						'szShutdown_cmd'				=>	"",
						'bInclude_Restart'				=>	1,
						'szRestart_cmd'					=>	"",
						'bInclude_Suspend'				=>	1,
						'bInclude_Suspend2SN' 			=>	1,
						'szSuspend_cmd'					=>	"",
						'bInclude_Hibernate'			=>	1,
						'bInclude_Hibernate2SN'			=>	1,
						'szHibernate_cmd'				=>	"",
						'bInclude_SCRestart'			=>	0,
						'szSCRestart_cmd'				=>	"",
						'bInclude_WebInterface'			=>	1,
						'bPowerOffPlayers'				=>	1,
						'nRegretDelay'					=>	15,
						'nGoToSNDelay'					=>	60,
						'bNoWebUIRedirect2SN',			=>	0,
						'bGotoSNAllPlayers'				=>	0,
						'szPushSNPlayerMACs'			=>	"",
						'bSetRTCWakeForAlarm'			=>	1,
						'szSetRTCWake_cmd'				=>	"",
						'nRTCWakeupAdvance'				=>	5,
						'bRetrievePlayersFromSN'		=>	0,
						'nRetrievePlayersFromSNWait'	=>	15,
						'szPullSNPlayerMACs'			=>	"",
						'bHookSleepButton'				=>	0,
						'szSleepButtonAction'			=>	'plugin_menu',
						'bUseSleepEndWatchdog'			=>	0,
						'szSleepEndWatchdogAction'		=>  'suspend',
						'bUseIdleWatchdog'				=>	0,
						'nIdleWatchdogTimeout'			=>	45,
						'szIdleWatchdogAction'			=>	'suspend',
						'bUseEODWatchdog'				=>	0,
						'nEODWatchdogTimeout',			=>	10,
						'szEODWatchdogStartTime'		=>	"00:30",
						'szEODWatchdogEndTime'			=>	"05:00",
						'szEODWatchdogAction'			=>	'shutdown',
						'szEODWatchdog_cmd'				=>	"",
						'bUseOnWakeupWatchdog'			=>	0,
						'szOnWakup_cmd'					=>	"",
		     );



#if (!defined( $g{prefs} )) {
#	$g{prefs} = preferences('plugin.srvrpowerctrl');
#	$g{log}->debug("Prefs initialized..");
#}


my $plugin; # the main plugin class

sub new {
	my $class = shift;
	$plugin   = shift;


	#Initialize log..
	InitLog();
	$g{log}->debug("Running on $g{szOS}");

	#Initialize prefs
	InitPrefs();

	#Report our prefs..
	DispPrefs();

	$class->SUPER::new;
}


sub InitLog {
	if (!defined( $g{log} )) {
		$g{log} = Slim::Utils::Log->addLogCategory({
			'category'     => 'plugin.SrvrPowerCtrl',
			'defaultLevel' => 'INFO',
			'description'  => 'PLUGIN_SRVRPOWERCTRL_MODULE_NAME',
			});
		$g{log}->debug("Log initialized..");
	}
}

sub InitPrefs {
	$g{log}->debug("Initializing prefs..");

	if (!defined( $g{prefs} )) {
		$g{prefs} = preferences('plugin.srvrpowerctrl');
		$g{prefs}->migrate(1, \&MigratePrefs);
	} else {
		$g{log}->debug("Prefs already defined!");
	}

	#Regularize our prefs..
	$g{log}->debug("Regularizing Prefs..");
	FixUpPrefs();

	#why is this necessary?
	foreach my $pref (@prefNames) {
		$g{prefs}->{$pref} = $g{prefs}->get($pref);
	}

	#set our prefs change handler..
	$g{log}->debug("Setting Prefs change handler..");
	$g{prefs}->setChange(\&PrefsChange, @prefNames);

	$g{log}->debug("Prefs initialized..");

	return defined($g{prefs});
}


sub MigratePrefs {

	$g{log}->debug("Migrating prefs..");

	#Check for stale prefs..reset to defaults if necessary..
	#If the plugin version is > than the prefs file version, reset all the prefs to defaults.
	if ( $g{prefs}->nPrefsVersion < $g{nVersion} ) {
		SetPrefDefaults();
	}

	$g{log}->debug("Prefs migrated..");
	return 1;
}


sub GetScriptPath {
	#from KDF's Execute Script..
	my $scriptPath = catfile((Slim::Utils::Prefs::dir() || Slim::Utils::OSDetect::dirsFor('prefs')),'scripts');
	return $scriptPath;
}



sub SetPrefDefaults {
	my $cmd;
	my $scriptdir;
	my $screstart;
	my $scwakeup;
	my $szVersion = $::VERSION;

	$g{log}->info("Setting prefs to $g{nVersion} defaults");
	$g{prefs}->set('nPrefsVersion', $g{nVersion} );

	$g{log}->info("Detected OS == $g{szOS}");


	#Get the default values..
	foreach my $pref (@prefNames) {
		$g{prefs}->set( $pref, $prefDefaults{$pref} );
	}

	$scriptdir = GetScriptPath();
	$g{log}->debug("Script Path == $scriptdir");


	$screstart = $scriptdir;

	if (length($szVersion) > 3) {
		#truncate the version: e.g. '7.2.1' becomes '7.2'
	       $szVersion = substr($szVersion, 0, 3);
		}


	##OS types can be 'unix', 'win' or 'mac'

	## Are we running windows?
  	if ($g{szOS} eq 'win') {
		$cmd = "c:\\windows\\system32\\cmd.exe";

		$screstart = $screstart . "\\screstart.cmd";
		$screstart = Win32::GetShortPathName( $screstart );

		$g{prefs}->set('szShutdown_cmd',	"$cmd /C start /B scpowertool.exe --shutdown --log" );
		$g{prefs}->set('szRestart_cmd',		"$cmd /C start /B scpowertool.exe --restart --log" );
		$g{prefs}->set('szSuspend_cmd',		"$cmd /C start /B scpowertool.exe --standby --log" );
		$g{prefs}->set('szHibernate_cmd', 	"$cmd /C start /B scpowertool.exe --hibernate --log" );
		$g{prefs}->set('szSCRestart_cmd',	"$cmd /C start /B starthidden.exe scpowertool.exe --screstart --log" );
		$g{prefs}->set('szSetRTCWake_cmd',	"$cmd /C start /B scpowertool.exe --wakeup=%d --quiet --log" );
		#$g{prefs}->set('szSetRTCWake_cmd',	"$cmd /C start /B wosb.exe /run /systray /ami dt=%f tm=%t" );

		$g{prefs}->set('bUseOnWakeupWatchdog',	1 );
		$g{prefs}->set('szOnWakup_cmd',		"$cmd /C start /B scpowertool.exe --keep-alive=20 --quiet --log" );

	} elsif ($g{szOS} eq 'unix') {
		$cmd = "sudo";

		$screstart = "/usr/local/sbin/screstart.sh";
		$scwakeup  = "/usr/local/sbin/scwakeup.sh";

		$g{prefs}->set('szShutdown_cmd',	"$cmd /sbin/shutdown -h now" );
		$g{prefs}->set('szRestart_cmd',		"$cmd /sbin/shutdown -r now" );
		$g{prefs}->set('szSuspend_cmd',		"$cmd /usr/sbin/pm-suspend" );
		$g{prefs}->set('szHibernate_cmd', 	"$cmd /usr/sbin/pm-hibernate" );
		$g{prefs}->set('szSCRestart_cmd',	"$cmd $screstart" );
		$g{prefs}->set('szSetRTCWake_cmd',	"$cmd $scwakeup %d" );

		$g{prefs}->set('szOnWakup_cmd',	"" );

	} elsif ($g{szOS} eq 'mac') {

		$scriptdir = '/Library/Scripts/SrvrPowerCtrl/';

		$g{prefs}->set('szShutdown_cmd',	"$scriptdir/srvrpowerctrl_helper.sh shutdown" );
		$g{prefs}->set('szRestart_cmd',		"$scriptdir/srvrpowerctrl_helper.sh restart" );
		$g{prefs}->set('szSuspend_cmd',		"$scriptdir/srvrpowerctrl_helper.sh suspend" );
		$g{prefs}->set('szHibernate_cmd', 	"$scriptdir/srvrpowerctrl_helper.sh hibernate" );
		$g{prefs}->set('szSCRestart_cmd',	"$scriptdir/srvrpowerctrl_helper.sh screstart"  );
		$g{prefs}->set('szSetRTCWake_cmd',	"$scriptdir/srvrpowerctrl_helper.sh scwakeup %d" );

	}


}


sub FixUpPrefs {
	foreach my $pref (@prefNames) {
		if ( !defined($g{prefs}->$pref) ) {
			#Don't set string prefs to 0
			if ( substr($pref,0,2) eq 'sz' ) {
				$g{prefs}->set($pref, "");
			} else {
				$g{prefs}->set($pref, 0);
			}
		} elsif ( $g{prefs}->$pref eq 'on' ) {
			$g{prefs}->set($pref, 1);
		}
	}
}

sub DispPrefs {
	my $showpref = shift;
	my $msg = shift;
	my @pnames;
	my $num = 0;
	my $preflabel;

	if (defined($showpref)) {
		push (@pnames, $showpref);
	} else {
		@pnames = @prefNames;
	}

	if (!defined($msg)) {
		$msg = " ";
	} else {
		$msg = $msg . " ";
	}

	#get the length of the longest prefname
	foreach my $pref (@prefNames) {
		if ($num < length($pref)) {
			$num = length($pref);
		}
	}
	foreach my $pref (@pnames) {
		$preflabel = sprintf('%*2$s', $pref, $num);
		$g{log}->debug("$msg $preflabel == $g{prefs}->{$pref}");
	}
}


# Prefs change..
# start using routines form Slim::Utils::Validate.pm to validate values: isInt(), numer(), isTime(), etc.

sub PrefsChange {
	my ($pref, $value) = @_;

	#validate boolian prefs
	if ( substr($pref, 0, 1) eq 'b' ) {
		if (!defined($value)) {
			#fix up the reference to silence a warning in the log..
			@_[1] = 0;
			$value = 0;
			$g{prefs}->{$pref} = $value;
		} elsif ($value eq 'on') {
			@_[1] = 1;
			$value = 1;
			$g{prefs}->{$pref} = $value;
		}
	#validate numeric prefs: if the value is not a number, reset it to the default..
	} elsif ( substr($pref, 0, 1) eq 'n' ) {
		if (!Plugins::SrvrPowerCtrl::Util::IsNumeric($value)) {
			$value = $prefDefaults{$pref};
			@_[1] = $value;
			$g{prefs}->{$pref} = $value;
		}
	#define any other undefined values..
	} else {
		if (!defined($value)) {
			#fix up the reference to silence a warning in the log..
			$value = "";
			@_[1] = $value;
			$g{prefs}->{$pref} = $value;
		}
	}

	#special validations...
	if ( ($pref eq 'szEODWatchdogStartTime') && (!defined(Slim::Utils::Validate::isTime($value))) ) {
		$g{log}->error("$value is not a valid time string!");
		$value = $prefDefaults{$pref};
		@_[1] = $value;
		$g{prefs}->{$pref} = $value;
		$g{prefs}->set($pref, $value);
	}

	if ( ($pref eq 'szEODWatchdogEndTime') && (!defined(Slim::Utils::Validate::isTime($value))) ) {
		$g{log}->error("$value is not a valid time string!");
		$value = $prefDefaults{$pref};
		@_[1] = $value;
		$g{prefs}->{$pref} = $value;
		$g{prefs}->set($pref, $value);
	}


	$g{log}->debug("PrefsChange: $pref == $value");

	#Actions to take on prefs change..
	if ( $pref eq 'bInclude_WebInterface' ) {
		Plugins::SrvrPowerCtrl::WebUI::ActivateWebUI($value);
	} elsif ( $pref eq 'bHookSleepButton' ) {
		Plugins::SrvrPowerCtrl::SleepButton::HookSleepButton($value);
		Plugins::SrvrPowerCtrl::SleepButton::ChangeSleepHoldButtonHandlers($value);
	} elsif ( $pref eq 'bUseSleepEndWatchdog' ) {
		Plugins::SrvrPowerCtrl::Watchdog::ActivateSleepEndWatchdog($value);
	} elsif ( $pref eq 'bUseIdleWatchdog' ) {
		Plugins::SrvrPowerCtrl::Watchdog::ActivateIdleWatchdog($value);
	} elsif ( $pref eq 'bUseEODWatchdog' ) {
		Plugins::SrvrPowerCtrl::Watchdog::ActivateEODWatchdog($value);
	} elsif ( $pref eq 'szEODWatchdogStartTime' || $pref eq 'szEODWatchdogEndTime') {
		Plugins::SrvrPowerCtrl::Watchdog::ActivateIdleWatchdog($value);
	} elsif ( $pref eq 'bRetrievePlayersFromSN' ) {
		#We need to enable the OnWakeupWatchdog here too..
		if ($value  && (!$g{prefs}->bUseOnWakeupWatchdog) ) {
			$g{prefs}->set('bUseOnWakeupWatchdog',	1 );
		}
	} elsif ( $pref eq 'bUseOnWakeupWatchdog' ) {
		Plugins::SrvrPowerCtrl::Watchdog::ActivateOnWakeupWatchdog($value);
		#Don't allow this pref to be disabled if bRetrievePlayersFromSN is active..
		if ($g{prefs}->bRetrievePlayersFromSN)  {
			$g{prefs}->set('bUseOnWakeupWatchdog',	1 );
		}
	} elsif ( $pref eq 'bSetRTCWakeForAlarm' ) {
		#Refresh the menu structure...prefs info gets pulled into the menu structure...
		Plugins::SrvrPowerCtrl::Menu::initActionItems();

	} elsif ( $pref eq 'restore_defaults' ) {
		$g{log}->debug("Attempting to restore defaults!");
		SetDefaults();
	}

	#DispPrefs($pref, "PrefsChange:");

	#reinitialize the action items..
	#if ($pref =~ m/^bInclude_/) {
		#Refresh the menu structure..
		#Plugins::SrvrPowerCtrl::Menu::initActionItems();
	#}

	Plugins::SrvrPowerCtrl::Menu::initActionItems();

	return $value;
}


#This adds the Name to the Settings->Advanced combobox.
#sub name {
#	return Slim::Web::HTTP::protectName('PLUGIN_SRVRPOWERCTRL_BASIC_SETTINGS');
#}

sub page {
	#$g{log}->is_debug && Plugins::SrvrPowerCtrl::Util::logArgs(@_);
	my $urlSettings = 'plugins/SrvrPowerCtrl/settings/basic.html';

	#$urlSettings = Plugins::SrvrPowerCtrl::WebUI::EscapeURL($urlSettings);

	return Slim::Web::HTTP::protectURI($urlSettings);
}

sub prefs {
	return ($g{prefs}, @prefNames );
}

sub handler {
	my ($class, $client, $params) = @_;
	my $action;
	my $item;

	if ($params->{'reset'}) {
		SetPrefDefaults();
	}

	return $class->SUPER::handler($client, $params);
}

sub needsClient {
	return 0;
}

sub getPrefs {
	return $g{prefs};
}



1;

__END__
