# ===============================================================================================================
#    SrvrPowerCtrl - a plugin for SqueezeCenter 7.1
#      Allows shutdown/restart/suspend/hibernation of your SqueezeCenter server
#      hardware via your SqueezeBox's IR remote or via a SBC
#
#    Version 7.3-RC8-20090505
#
#    Copyright (C) 2008 Gordon Harris
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of the GNU General Public License,
#    version 2, which should be included with this software.
#
#    Based on Adrian C's (aka tenwiseman) Server Power Control plugin: http://adrianonline.wordpress.com/
#                                                   http://forums.slimdevices.com/showthread.php?t=32674
#    Copyright AdrianOnline.Net Feb 2007. Initially modeled on Max Spicer's Max2.pm
#
#    Also includes code from PeterW's AllQuiet plugin: http://www.tux.org/~peterw/slim/slim7/AllQuiet/
#    AllQuiet copyright (c) 2007 by Peter Watkins (peterw@tux.org)
#
#
#    See a duscussion of this plugin at http://forums.slimdevices.com/showthread.php?t=48521
#
# ===============================================================================================================
#
#    Util.pm -- Misc routines..
#

package Plugins::SrvrPowerCtrl::Util;

use base qw(Slim::Plugin::Base);
use strict;

#Global Variables..
use vars qw(%g);
*g = \%Plugins::SrvrPowerCtrl::Settings::g;

use Slim::Utils::Strings qw(string);


#sub max ($$) { $_[$_[0] < $_[1]] }
#
#sub min ($$) { $_[$_[0] > $_[1]] }


sub IsNumeric {
	my ($num) = @_;

	#$g{log}->debug("Testing $num");

	if (!defined($num)) {
		return 0;
 	} elsif ($num =~ /^-?\d/) {
		return 1;
	}
	return 0;
}

#does not work!
#sub IsCoderef {
#	$g{log}->debug(@_ . " is a " . ref(@_) );
#
#	if (ref(@_) eq 'CODE') {
#		return 1;
#	}
#	return 0;
#}

sub AnyPlayersPlaying {
	my @clients;
	my $curclient;

	@clients = Slim::Player::Client::clients();

	foreach $curclient (@clients) {
		#$g{log}->debug ($curclient->name() . " playmode is \[" . $curclient->playmode() . "\]");
		if (Slim::Player::Source::playmode($curclient) eq 'play') {
			return 1;
		}
	}
	return 0;
}

sub AnyPlayersUpdating {
	my @clients;
	my $curclient;

	@clients = Slim::Player::Client::clients();

	foreach $curclient (@clients) {
		#Just block on isUpgrading, not on needsUpgrade..
		if ( $curclient->isUpgrading() ) {
			$g{log}->info ($curclient->name() . " needsUpgrade() is \[" . $curclient->needsUpgrade() . "\]");
			return 1;
		}
	}
	return 0;
}


sub DisplayPlayerMessage {
	my ($client, $message, $duration, $skipclient, $jivemessage) = @_;
	my @clients;
	my $curclient;
	my $numclients;

	if (!defined($duration)) {
		$duration = $g{prefs}->nRegretDelay;
	}

	#$g{log}->debug("     client: $client");
	#$g{log}->debug("    message: $message");
	#$g{log}->debug("   duration: $duration");
	#$g{log}->debug(" skipclient: $skipclient");
	#$g{log}->debug("jivemessage: $jivemessage");

	if (defined($client)) {
		push(@clients, $client);
	} else {
		@clients = Slim::Player::Client::clients();
	}

	$numclients = @clients;

	#fixup the message..change underscores to spaces..
	$message =~ s/_/ /sig;

	#$g{log}->debug( "Displaying message \[" . $message . "\] on " . $numclients . " clients for " . $duration . " seconds.");

	foreach $curclient (@clients) {
		#Don't display a message here..
		if ( defined($skipclient) && ($skipclient->id() eq $curclient->id()) ) {
			#$g{log}->debug( "Not displaying message on " . $curclient->name() );
			next;
		}

		#Don't bother trying to disply the message on a SBR or SBC..
		#if ($curclient->deviceid eq 7 || $curclient->deviceid eq 9) {
		if ($curclient->deviceid eq 7) {
			#$g{log}->debug("Skipping display on ". $curclient->name() );
			next;
		}

		#$g{log}->debug( "Displaying message on " . $curclient->name() );
		#Change the font on the BOOM:
		if ($curclient->deviceid == 10) {
			$curclient->showBriefly(
				{
				'center' => [ '', '0' ],
				'fonts' => {	'graphic-160x32' => { 'center' => [ 'standard_n.1', 'logoSB2.2' ] }, },
				'line1' => string('PLUGIN_SRVRPOWERCTRL_MODULE_NAME'),
				'line2' => $message
				},
				{
					'duration' => "$duration"
				} );

		} else {
			#$g{log}->debug("Displaying $message on ". $curclient->name() . "..");
			$curclient->showBriefly(
				{
					'line1' => string('PLUGIN_SRVRPOWERCTRL_MODULE_NAME'),
					'line2' => $message
				},
				{
					'duration' => "$duration"
				} );
		}

		#show on the SBC too..
		if (defined($jivemessage)) {
			$curclient->showBriefly( { 'jive' => { 'text'    => [ $jivemessage ], } },{'duration' => $duration, 'block' => 0, } );
		} else {
			$curclient->showBriefly( { 'jive' => { 'text'    => [ $message ], } },{'duration' => $duration, 'block' => 0, } );
		}

	}
}

sub BlockPlayerMessage {
	my ($client, $message) = @_;
	my @clients;
	my $curclient;
	my $line1;
	my $line2;

	if (defined($client)) {
		push(@clients, $client);
	} else {
		@clients = Slim::Player::Client::clients();
	}

	#fixup the message..change underscores to spaces..
	$message =~ s/_/ /sig;

	$g{log}->debug( ($client ? $client->name() : "All Clients") . ": " . ($message ? $message : "no message.") );

	foreach $curclient (@clients) {

		#show on the SBC too..can't block here as it will block the controlled player too..
		$curclient->showBriefly( { 'jive' => { 'text'    => [ $message ], } },{'duration' => "15", 'block' => 0, } );

		$curclient->showBriefly(	{ 'line1' => string('PLUGIN_SRVRPOWERCTRL_MODULE_NAME'), 'line2' => $message },
									{	'duration' => "15",
										#'brightness' => "4",
										'block' => 1 },  );

		#$curclient->block( string('PLUGIN_SRVRPOWERCTRL_MODULE_NAME'), $message);
	}

}

sub UnBlockPlayerMessage {
	my ($client, @args) = @_;
	my $message = $args[0];
	my $duration = $args[1];
	my @clients;
	my $curclient;

	$g{log}->debug( ($client ? $client->name() : "All Clients") . ": " . ($message ? $message : "no message.") );

	if (defined($client)) {
		push(@clients, $client);
	} else {
		@clients = Slim::Player::Client::clients();
	}

	foreach $curclient (@clients) {
		$curclient->unblock();
		$curclient->showBriefly( { 'jive' => { 'text'    => [ " " ], } },{'duration' => 10, 'block' => 0, } );
	}

	if (defined($message)) {
		$g{log}->debug($message);
		DisplayPlayerMessage($client, $message, $duration);
	}

}

sub PowerOffPlayer {
	my ($client) = @_;
	my @clients;
	my $curclient;

	if (defined($client)) {
		push(@clients, $client);
	} else {
		@clients = Slim::Player::Client::clients();
	}

	foreach $curclient (@clients) {
		$g{log}->debug("Switching player " . $curclient->name() . " to OFF");
		$curclient->execute(['power', '0']);
		#$curclient->power(0);
		#experiment with turning the brightness to 0
		#$g{log}->debug( "Trying to turn the " . $curclient->name() . " display dark..");
		#How do I check to see that a client poseses "brightness"?
		#$curclient->brightness(0);
		#$curclient->execute(['show', 'line1:Dark%20Display', 'line2:Should%20not%20see%20me', 'duration:30', 'centered:1', 'brightness:0']);

	}
}

sub RestartSC {
	$g{log}->debug("Attempting SC7.4 style service restart..");
	Slim::Control::Request::executeRequest(undef, ['restartserver']);
}

sub GetWhenShouldRun {
	my @aTime;
	my $szTime;

	@aTime = localtime(time());

	#forward 2 minutes..at on linux seems to need this..
	if ($aTime[0] >= 30) {
		$aTime[1] += 2;
	} else {
		#forward 1 minute..
		$aTime[1] += 1;
	}

	#we don't need no stinking seconds..
	$aTime[0] = 0;

	#fixup minutes..
	if ($aTime[1] >= 60) {
		if ($aTime[2] >= 23) {
			$aTime[2] = 0;
		}
		$aTime[1] -= 60;
	}

	$szTime = sprintf("%02d:%02d", $aTime[2], $aTime[1]);

	return $szTime;
}

sub ScheduleCommand {
	my ($client, $item) = @_;

	#Find the next alarm and program the RTC to wake up the server at the appropriate time.
	if ($item->{'setrtcwakeup'}) {
		Plugins::SrvrPowerCtrl::Alarms::SetRTCWakeup();
	}

	#cue up our action command to execute in the future..
	return Slim::Utils::Timers::setTimer(undef, time() + $item->{'cmdwait'},
		sub {
			my $res;
			my $command;
			my $cmdargs;
			my @cmd;

			#We're the timer action...clear the timer ID since it's fired..
			$g{tPendingActionTimer} = 0;

			$command = $item->{command};

			# execute the coderef (if any) to get args for the cmd..
			if ( defined($item->{getcommandargcoderef}) ) {
				#execute the coderef..
				$cmdargs = &{$item->{getcommandargcoderef}}($client, $item);
				if ( defined($cmdargs) ) {
					if ( (IsNumeric($cmdargs)) && ($cmdargs == -1) ) {
						#fail mode..
						my $failmsg = sprintf( string('PLUGIN_SRVRPOWERCTRL_FAILURE_MSG'),  ucfirst( lc($item->{action}) ) );
						$g{log}->error("$item->{action} $item->{getcommandargcoderef} failed!");
						UnBlockPlayerMessage(undef, $failmsg);
						return;
					} else {
						#anything else returned is a cmdline arg for our script..
						$command = $item->{command} . ' ' . $cmdargs;
					}
				} else {
					#the coderef returned undef..
					$command = $item->{command};
				}
			}

			#if the command is a coderef rather than a script or system command name, execute it..
			if (ref($item->{command}) eq 'CODE') {
				$g{log}->debug("Command is a coderef!");
				#execute the coderef..
				$cmdargs = &{$item->{command}}($client, $item);
			#else the command is a system command or script..
			} else {
				#Embed the 'whenshouldrun' time to the args only if there is a '%s' sprintf formatting flag present.
				if (index($command, "%s") > 0) {
					$command = sprintf($command, GetWhenShouldRun());
				}

				$g{log}->info("Now trying to execute $command");
				@cmd = split(' ', $command );
				$res = system @cmd[0..$#cmd];
				#$res = "We are not doing this!";
				$g{log}->debug("Command returned $res");
				#$res == 1 may indicate a sudo permissions problem...not sure about return values from cmd.exe..

			}

			if ($item->{'stopsc'}) {
				if (defined($client)) {
					#from SqueezeTray.pl
					$client->execute(['stopserver']);
				} else {
					main::stopServer();
				}
			}

			if (!$item->{dispblock}) {
				UnBlockPlayerMessage();
			}

		}, );
}

sub PushToSqueezeNetwork {
	my ($client) = @_;
	my @clients;
	my $curclient;
	my $curclientname;
	my $macaddr;
	my $specificmacaddr;
	my $szSNPushedPlayers;
	my @aSNPushedPlayers = ();
	my $szNextAlarmClientMAC;
	my $szMACPushList;
	my @aMACPushList;
	my $nMACPushList = 0;

	#get the mac address of the next alarm owner..
	my ($nRTCWakeupTime, $szNextAlarmClientMAC) = Plugins::SrvrPowerCtrl::Alarms::GetNextAlarm();


	my $snserver = Slim::Networking::SqueezeNetwork->get_server('sn');
	@clients = ();

	#Get our list of specific MACs to restore from SN from the prefs..
	$szMACPushList = $g{prefs}->szPushSNPlayerMACs;
	#delimiters could be ; , or space
	@aMACPushList = split(/[\;\,\s]/, $szMACPushList );
	$nMACPushList = scalar @aMACPushList || 0;

	#Did we get an array of MACs to push from the prefs?
	if($nMACPushList) {
		$g{log}->debug( "szMACPushList: $szMACPushList == $nMACPushList players..");
	}

	#Are we pushing just a single client??
	if ( (!$g{prefs}->bGotoSNAllPlayers) && defined($client) ) {
		push( @clients, $client );
	} elsif ($nMACPushList) {
		@clients = Slim::Player::Client::clients();
	} else {
		@clients = Slim::Player::Client::clients();
	}

	#Are we pushing only specific players?
	if ( $nMACPushList > 0 ) {
		$g{log}->debug( "Trying to switch $nMACPushList players to SN..");
		foreach $specificmacaddr (@aMACPushList) {
			if (defined($specificmacaddr)) {
				foreach $curclient (@clients) {
					$macaddr = $curclient->id();
					if (defined($macaddr)) {
						if (lc($macaddr) eq lc($specificmacaddr)) {
							$curclientname = $curclient->name() || "unknown";
							$g{log}->debug( "Attempting to connect " . $curclientname . ": " . $curclient->id() . " to $snserver..");
							$curclient->execute(['connect', $snserver]);
							push( @aSNPushedPlayers, $curclient->id() );
						}
					}
				}
			}
		}
	} else {
		#push the players over to SN
		foreach $curclient (@clients) {
			$curclientname = $curclient->name() || "unknown";
			$g{log}->debug( "Attempting to connect " . $curclientname . ": " . $curclient->id() . " to $snserver..");
			$curclient->execute(['connect', $snserver]);
			push( @aSNPushedPlayers, $curclient->id() );
		}
	}

	#Check the mac of the next alarm..  If it's not in our list, then the player was already on SN.
	foreach my $szCurMAC (@aSNPushedPlayers) {
		if ( lc($szCurMAC) eq lc($szNextAlarmClientMAC) ) {
			$szNextAlarmClientMAC = undef;
		}
	}

	#Add that mac to our list of players to be pulled back on wakeup..
	if (defined($szNextAlarmClientMAC)) {
		$g{log}->debug( "Adding $szNextAlarmClientMAC to our list of pushed players because it has an alarm scheduled..");
		push( @aSNPushedPlayers, $szNextAlarmClientMAC );
	}

	#Save the list of pushed players in the prefs for later restore..
	$szSNPushedPlayers = join(';', @aSNPushedPlayers);
	$g{prefs}->set('szSNPushedPlayers',	$szSNPushedPlayers );

	return undef;
}

#Get an array of the players currently connected to SqueezeNetwork MACs..
sub GetSNPlayers {
	my @aSNPlayerMACs;
	my @aSNPlayers = Slim::Networking::SqueezeNetwork::Players->get_players();
	my $nSNPlayerCount = scalar @aSNPlayers  || 0;

	my $szDebugMsg = sprintf("nSNPlayerCount==%d", $nSNPlayerCount);
	$g{log}->debug ("$szDebugMsg");

	foreach my $player (@aSNPlayers ) {
		$g{log}->debug ("Player connected to SN: " . $player->{mac} . " - " . $player->{name} );
		push(@aSNPlayerMACs, $player->{mac});
	}

	return @aSNPlayerMACs;
}

sub PullFromSqueezeNetwork {
	my $szMACRestoreList;
	my @aMACRestoreList;
	my $nMACRestoreList = 0;
	my $request;
	my @aSNPlayerMACs;
	my $nSNPlayerMACs = 0;
	my $nDisconnectCount = 0;
	my $specificmacaddr;
	my $macaddr;
	my $nSNPushedPlayers;
	my $szSNPushedPlayers;
	my @aSNPushedPlayers = ();
	my $snserver = Slim::Networking::SqueezeNetwork->get_server("sn");

	if ( !$g{prefs}->bUseOnWakeupWatchdog ) {
		$g{log}->debug( "Need to activate OnWakeupWatchdog..");
		$g{prefs}->set('bUseOnWakeupWatchdog',	1 );
		Plugins::SrvrPowerCtrl::Watchdog::ActivateOnWakeupWatchdog(1);
	}

	#Get our list of specific MACs to restore from SN from the prefs..
	$szMACRestoreList = $g{prefs}->szPullSNPlayerMACs;
	#delimiters could be ; , or space
	@aMACRestoreList = split(/[\;\,\s]/, $szMACRestoreList );
	$nMACRestoreList = scalar @aMACRestoreList || 0;

	#Did we get an array of MACs to restore from the prefs?
	if($nMACRestoreList) {
		$g{log}->debug( "szMACRestoreList: $szMACRestoreList == $nMACRestoreList players..");
	}

	#Get our list of MACs that we previously pushed onto SN..
	$szSNPushedPlayers = $g{prefs}->szSNPushedPlayers;
	#delimiters could be ; , or space
	@aSNPushedPlayers = split(/[\;\,\s]/, $szSNPushedPlayers );
	$nSNPushedPlayers = scalar @aSNPushedPlayers || 0;
	if($nSNPushedPlayers) {
		$g{log}->debug( "szSNPushedPlayers: $szSNPushedPlayers == $nSNPushedPlayers players..");
	}

	#Get the players currently connected to SN..
	@aSNPlayerMACs = GetSNPlayers();
	$nSNPlayerMACs = scalar @aSNPlayerMACs || 0;

	#if we don't have any players connected to SN, give up..
	if ($nSNPlayerMACs == 0) {
		$g{log}->debug( "No players to restore..");
		#Erase the list of pushed players..
		$g{prefs}->set('szSNPushedPlayers',	'');
		return;
	}

	#Are we pulling only specific players?
	if ( $nMACRestoreList > 0 ) {
		$g{log}->debug( "Trying to restore only $nMACRestoreList players from SN..");
		foreach $specificmacaddr (@aMACRestoreList) {
			if (defined($specificmacaddr)) {
				$g{log}->debug( "Checking player $specificmacaddr to see if it's connected..");
				foreach $macaddr (@aSNPlayerMACs) {
					if (defined($macaddr)) {
						if (lc($macaddr) eq lc($specificmacaddr)) {
							$g{log}->debug( "Attempting to disconnect $macaddr from $snserver...");
							$request = Slim::Control::Request::executeRequest(undef, [ 'squeezenetwork', 'disconnect', $macaddr ]);
						}
					}
				}
			}
		}
	} else {
		$g{log}->debug( "Trying to restore ALL pushed players from SN..");

		if (!defined(@aSNPushedPlayers) ) {
			$g{log}->debug( "No pushed player list!");
			@aSNPushedPlayers = @aSNPlayerMACs;
		}
		#Restore only the players we pushed onto SN when we suspended..
		foreach $specificmacaddr (@aSNPushedPlayers) {
			if (defined($specificmacaddr)) {
				$g{log}->debug( "Checking player $specificmacaddr to see if it's connected..");
				foreach $macaddr (@aSNPlayerMACs) {
					if (defined($macaddr)) {
						if (lc($macaddr) eq lc($specificmacaddr)) {
							$g{log}->debug( "Attempting to disconnect $macaddr from $snserver...");
							$request = Slim::Control::Request::executeRequest(undef, [ 'squeezenetwork', 'disconnect', $macaddr ]);
						}
					}
				}
			}
		}
	}

	#Erase the list of pushed players..
	$g{prefs}->set('szSNPushedPlayers',	'');

}


sub TestCoderefFail {
	my ($client, $item) = @_;

	$g{log}->debug("Testing failure mode..");

	if (defined($client)) {
		$g{log}->debug("client == " . $client->name() . ", type == " . $client->deviceid );
	} else {
		$g{log}->debug("client == noclient");
	}

	$g{log}->debug("action == " . $item->{action} );

	return -1;
}

sub logHashElements {

	my @myhash = shift;
	my $num = 0;

	foreach my $item (@myhash) {
		$g{log}->debug("item[$num]:");
		while( my ($k, $v) = each %$item ) {
 			$g{log}->debug("key: $k, value: " . ( defined($v) ? $v : "undef" ) );
    		}
		$num++;
	}
}

sub logArgs {
	my (@args) = @_;
	my $n = 0;

	while (defined($args[$n])) {
		$g{log}->debug("arg[$n] == $args[$n++]");
	}
}

1;
